package net.sergeych.kiloparsec.adapter

import io.ktor.network.selector.*
import io.ktor.network.sockets.*
import io.ktor.utils.io.CancellationException
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.Flow
import net.sergeych.kiloparsec.KiloClient
import net.sergeych.kiloparsec.KiloServer
import net.sergeych.kiloparsec.RemoteInterface
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.tools.AtomicCounter

internal val udpCounter = AtomicCounter(0)

class UdpTransportException(override val message: String) : RemoteInterface.InvalidDataException(message)

/**
 * Listen for incoming UDP connections and provide transport flow for it. See also [UdpServer.transportFlow]
 * for another way to create a server. Use it with [KiloServer]:
 * ```kotlin
 * // Whatever server session data we might need:
 * data class Session(
 *     var data: String,
 * )
 *
 * // declare some commands (normally in a shared module):
 * val cmdSave by command<String, Unit>()
 * val cmdLoad by command<Unit, String>()
 * val cmdDrop by command<Unit, Unit>()
 * val cmdException by command<Unit, Unit>()
 *
 * // Interface using the session above, can be shared between many
 * // server types and instances (different ports and protocols):
 * val cli = KiloInterface<Session>().apply {
 *     onConnected { session.data = "start" }
 *     on(cmdSave) { session.data = it }
 *     on(cmdLoad) {
 *         session.data
 *     }
 *     on(cmdException) {
 *         throw TestException()
 *     }
 *     on(cmdDrop) {
 *         throw LocalInterface.BreakConnectionException()
 *     }
 * }
 * // Now create a server to accept incoming UDPs on our port:
 * val server = KiloServer(cli, acceptUdpDevice(port)) {
 *     // This initializes new session for each incoming command
 *     Session("unknown")
 * }
 * ```
 *
 * See [connectUdpDevice] for the client sample.
 *
 * When it is necessary to stop listening to some port, use [UdpServer] instead.
 */
fun acceptUdpDevice(port: Int, localInterface: String = "0.0.0.0"): Flow<InetTransportDevice> =
    UdpServer(port, localInterface).transportFlow

/**
 * Connect to UDP server (see [acceptUdpDevice]) and return a [InetTransportDevice] for it. It
 * should be used with [KiloClient] as connection provider:
 * ```kotlin
 * val client = KiloClient<Unit>() {
 *     connect { connectUdpDevice("localhost:$port") }
 * }
 * // now we can execute remote commands:
 * assertEquals("start", client.call(cmdLoad))
 * ```
 */
fun connectUdpDevice(hostPort: String) = connectUdpDevice(hostPort.toNetworkAddress())

/**
 * Connect to UDP server (see [acceptUdpDevice]) and return a [InetTransportDevice] for it. It
 * should be used with [KiloClient] as connection provider:
 * ```kotlin
 * val client = KiloClient<Unit>() {
 *     connect { connectUdpDevice("localhost:$port") }
 * }
 * // now we can execute remote commands:
 * assertEquals("start", client.call(cmdLoad))
 * ```
 */
fun connectUdpDevice(addr: NetworkAddress): InetTransportDevice {
    val selectorManager = SelectorManager(Dispatchers.IO)
    val remoteAddress = InetSocketAddress(addr.host, addr.port)
    val socket = aSocket(selectorManager).udp().connect(remoteAddress)

    val done = CompletableDeferred<Unit>()

    val transport = UdpSocketTransport(object : UdpConnector {
        override suspend fun sendBlock(block: UdpBlock, toAddress: SocketAddress) {
            socket.send(block.toDatagram(remoteAddress))
        }

        override suspend fun disconnectClient(address: SocketAddress) {
            done.complete(Unit)
        }

    }, remoteAddress, false)

    globalLaunch {
        launch {
            while (isActive) {
                try {
                    transport.processIncoming(UdpBlock.decode(socket.receive()))
                } catch (_: CancellationException) {
                    break
                } catch (e: Exception) {
                    transport.close()
                    break
                }
            }
        }
        done.await()
    }

    return transport.transportDevice
}